@@ -1,16 +1,14 @@
 import os
+from datetime import datetime
+from enum import Enum
+
+import pandas as pd
+from fuzzywuzzy import fuzz
 from groclient import GroClient
 
-
-from grocropclient.constants import FrequencyList, OtherItemsList, SourceList, MetricList, CropList, GRO_YIELD_MODELS
-
-from fuzzywuzzy import fuzz
-
 from grocropclient import crop_budgets
+from grocropclient.constants import FrequencyList, OtherItemsList, SourceList, MetricList, CropList, GRO_YIELD_MODELS, AreaWeightingSeriesList, ItemNameCategoryList, CROP_BUDGET_REGIONS
 
-from typing_extensions import Literal
-from enum import Enum
-from datetime import datetime
 
 API_HOST = "api.gro-intelligence.com"
 
@@ -21,14 +19,17 @@ AVAILABLE_ITEMS   = [s.value for s in CropList] + [s.value for s in OtherItemsLi
 AVAILABLE_METRICS_NAMES = [(s.value, s.name.replace('_', ' ')) for s in MetricList]
 AVAILABLE_ITEMS_NAMES   = [(s.value, s.name.replace('_', ' ')) for s in CropList] + [(s.value, s.name.replace('_', ' ')) for s in OtherItemsList]
 
-# Returns True if the source/item/metric selection is within the US Crop tier.
-def is_valid_selection(dataseries):
+
+
+def is_valid_selection(dataseries: dict) -> bool:
+    """Returns True if the source/item/metric selection is within the US Crop tier."""
     return dataseries['source_id'] in AVAILABLE_SOURCES and \
     dataseries['metric_id'] in AVAILABLE_METRICS and \
     dataseries['item_id'] in AVAILABLE_ITEMS
 
-# Return the number of days between today's data and a Gro date.
-def days_ago(gro_date):
+
+def days_ago(gro_date: str):
+    """Return the number of days between today's data and a Gro date."""
     # Wer remove the TZ ([0:10]) before we parse.
     return (datetime.today() - datetime.strptime(gro_date[0:10], "%Y-%m-%d")).days
 
@@ -43,12 +44,32 @@ def handle_enums(obj):
         obj[k] = list(map(fix_enum, v)) if isinstance(v, list) else fix_enum(v)
   elif isinstance(obj, list):
         obj = [fix_enum(v) for v in obj]
-  elif isinstance(obj, Enum):
-        obj = fix_enum(obj)
   else:
-      raise Exception("Couldn't handle the following with handle_enums: "+str(obj))
+        obj = fix_enum(obj)
+  
   return obj
 
+def get_region_id(region_name: str, source_name: str) -> int:
+    """Map crop budget region_name to numeric region_id"""
+    region_name = region_name.lower()
+    source_name = source_name.lower()
+    return CROP_BUDGET_REGIONS[source_name][region_name]
+
+
+def is_corn_or_soybeans(item_name: str) -> str:
+    """
+    This method is required because a different series exists for either corn or soybeans for Yield and Price 
+    in the Area Weighting service.
+    
+    Returns
+    -------
+    either 'corn' or 'soybeans'
+    """
+    for word in item_name.split(' '):
+        if ItemNameCategoryList.CORN.value in word.lower():
+            return ItemNameCategoryList.CORN.value
+        elif ItemNameCategoryList.SOYBEANS.value in word.lower():
+            return ItemNameCategoryList.SOYBEANS.value
 
 class GroCropClient(object):
 
@@ -135,6 +156,67 @@ class GroCropClient(object):
         """
         return crop_budgets.get_crop_budget_as_df(self._api_host, self._access_token, source_name=source_name, item_name=item_name, region_name=region_name, productivity=productivity)
 
+    def get_crop_budget_yield_model(self, item_name: str, region_name: str, source_name: str, latest_date_only=False) -> pd.DataFrame:
+        """
+        Get the Gro Yield Model values (bu/acre) for the given item and region combination. If `latest_date_only` is True, only get the latest Yield value.
+        The Yield value returned for each date is a weighted average of the district region values within the requested region,
+        where the area of Corn or Soybeans harvested in those districts represents the weight.
+
+        Parameters
+        ----------
+        item_name: str
+        The `item_name` is from the result of from the result of get_crop_budgets(…)
+        region_name: str 
+        The `region_name` is from the result of from the result of get_crop_budgets(…)
+        latest_date_only: bool
+        If `latest_date_only` is True, only get the latest Yield value.
+
+        Returns 
+        -------
+                    Yield 
+        2000-02-25  0.217
+        2000-03-04  0.217
+        2000-03-12  0.221
+        """
+        region_id = get_region_id(region_name, source_name)
+        crop = is_corn_or_soybeans(item_name)
+        if crop == ItemNameCategoryList.CORN.value:
+            yield_dict = self._client.get_area_weighted_series(AreaWeightingSeriesList.CORN_YIELD.value, ['NASS Corn (US only)'], region_id, latest_date_only)
+        elif crop == ItemNameCategoryList.SOYBEANS.value:
+            yield_dict = self._client.get_area_weighted_series(AreaWeightingSeriesList.SOYBEANS_YIELD.value, ['NASS Soybeans (US only)'], region_id, latest_date_only)
+ 
+        return pd.DataFrame({'Yield': yield_dict.values()}, yield_dict.keys())
+    
+
+    def get_crop_budget_price_model(self, item_name: str, region_name: str, source_name: str, latest_date_only=False) -> pd.DataFrame:
+        """
+        Get the DTN Price ($/bu) values for the given item and region combination. If `latest_date_only` is True, only get the latest Price value.
+        The price value returned for each date is an average of the district region values within the requested region.
+
+        Parameters
+        ----------
+        item_name: str
+        The `item_name` is from the result of from the result of get_crop_budgets(…)
+        region_name: str 
+        The `region_name` is from the result of from the result of get_crop_budgets(…)
+        latest_date_only: bool
+        If `latest_date_only` is True, only get the latest Price value.
+
+        Returns 
+        -------
+                    Price
+        2000-02-25  0.217
+        2000-03-04  0.217
+        2000-03-12  0.221
+        """
+        region_id = get_region_id(region_name, source_name)
+        crop = is_corn_or_soybeans(item_name)
+        if crop == ItemNameCategoryList.CORN.value:
+            price_dict = self._client.get_area_weighted_series(AreaWeightingSeriesList.CORN_PRICE.value, ['AVERAGE'], region_id,latest_date_only)
+        elif crop == ItemNameCategoryList.SOYBEANS.value:
+            price_dict = self._client.get_area_weighted_series(AreaWeightingSeriesList.SOYBEANS_PRICE.value, ['AVERAGE'], region_id, latest_date_only)
+
+        return pd.DataFrame({'Price': price_dict.values()}, price_dict.keys())
     # END OF CROP BUDGET functions
 
     def lookup(self, entity_type, entity_ids):
@@ -201,7 +283,6 @@ class GroCropClient(object):
             for i in range(5):
             print(next(dataseries_gen))
         """
-
         dataseries_gen = self._client.find_data_series(**kwargs)
         while True:
             result = next(dataseries_gen)
@@ -238,7 +319,6 @@ class GroCropClient(object):
         kwargs = handle_enums(kwargs)
 
         dataseries = self._client.get_data_series(**kwargs)
-
         filtered_dataseries=[series for series in dataseries if is_valid_selection(series)]
         return filtered_dataseries
 
@@ -413,9 +493,9 @@ class GroCropClient(object):
             ranked_results = sorted( [ (k[0], k[1], fuzz.token_set_ratio(search_terms, k[1])) for k in AVAILABLE_METRICS_NAMES], key=lambda x:x[2], reverse=True)
             for result in ranked_results:
                 yield { 'metric_id': result[0], 'metric_shortname': result[1], 'score': result[2] }
-
+        
         if entity_type == 'items':
-            ranked_results = sorted([(k[0], k[1], fuzz.token_set_ratio(search_terms, k[1])) for k in AVAILABLE_METRICS_NAMES], key=lambda x:x[2], reverse=True)
+            ranked_results = sorted([(k[0], k[1], fuzz.token_set_ratio(search_terms, k[1])) for k in AVAILABLE_ITEMS_NAMES], key=lambda x:x[2], reverse=True)
             for result in ranked_results:
                 yield { 'item_id': result[0], 'item_shortname': result[1], 'score': result[2] }
 
@@ -454,7 +534,7 @@ class GroCropClient(object):
         """
         self._client._data_series_list = set()
         self._client._data_series_queue = []
-        self._client._data_frame = pandas.DataFrame()
+        self._client._data_frame = pd.DataFrame()
 
     def get_df(self, **kwargs):
         """Call :meth:`~.get_data_points` for each saved data series and return as a combined
@@ -475,7 +555,7 @@ class GroCropClient(object):
     ## Yield ##
     ###########
 
-    def get_all_yield_models():
+    def get_all_yield_models(self):
         """Returns all yield models as an array of {"id": …, "name": …}. """
         return [{"id": k, "name": "%s / %s" % (v["region_name"], v["item_name"])} for k,v, in GRO_YIELD_MODELS.items()]
 
@@ -502,18 +582,29 @@ class GroCropClient(object):
         if len(yield_model_data) == 0:
             print("no data")
             return None
-        unit_id = yield_model_data[0]['unit_id']
+        for point in yield_model_data:
+            unit_id = point['unit_id']
+            if unit_id != None:
+                break
         unit_abbrev = self.lookup_unit_abbreviation(unit_id)
-        return [ { "date": k["available_date"][0:10],
-                   "value": "%f ± %f (%s)" % (k["value"],
-                   k["metadata"]["conf_interval"], unit_abbrev)} for k in yield_model_data \
-                if days_ago(k["available_date"]) < N][::-1]
+        summary = [] 
+        for k in yield_model_data:
+            if days_ago(k["available_date"]) < N:
+                k_summary = {"date": k["available_date"][0:10]}
+                if ('metadata' in k) & ("conf_interval" in k['metadata']):
+                    k_summary["value"] =  "%f ± %f (%s)" % (k["value"],
+                       k["metadata"]["conf_interval"], unit_abbrev)
+                else:
+                    k_summary["value"] =  "%f (%s)" % (k["value"], unit_abbrev)
+                summary.append(k_summary)
+        summary = summary[::-1]
+        return summary
 
     ############
     ## Prices ##
     ############
 
-    def get_cash_prices_spot_county_level(self, crop_id, county_id, start_date = None, end_date = None):
+    def get_cash_prices_spot_county_level(self, crop_id: int, county_id: int, start_date: str = None, end_date: str = None):
         """Returns cash prices for a crop and a county."""
         params = { "source_id": SourceList.DTN_DISTRICT_AGGREGATED,
                    "frequency_id": 1,
@@ -525,7 +616,7 @@ class GroCropClient(object):
         }
         return self.get_data_points(**params)
 
-    def get_cash_prices_new_crop_county_level(self, crop_id, county_id, start_date = None, end_date = None):
+    def get_cash_prices_new_crop_county_level(self, crop_id: int, county_id: int, start_date: str = None, end_date: str = None):
         """Returns cash prices for a crop and a county."""
         params = { "source_id": SourceList.DTN_DISTRICT_AGGREGATED,
                    "frequency_id": 1,
@@ -537,7 +628,7 @@ class GroCropClient(object):
         return self.get_data_points(**params)
 
 
-    def get_basis_prices_spot_county_level(self, crop_id, county_id, start_date = None, end_date = None):
+    def get_basis_prices_spot_county_level(self, crop_id: int, county_id: int, start_date: str = None, end_date: str = None):
         """Returns basis prices for a crop and a county."""
         params = { "source_id": SourceList.DTN_DISTRICT_AGGREGATED,
             "frequency_id": 1,
@@ -549,7 +640,7 @@ class GroCropClient(object):
            }
         return self.get_data_points(**params)
 
-    def get_basis_prices_new_crop_county_level(self, crop_id, county_id, start_date = None, end_date = None):
+    def get_basis_prices_new_crop_county_level(self, crop_id: int, county_id: int, start_date: str = None, end_date: str = None):
         """Returns basis prices for a crop and a county."""
         params = { "source_id": SourceList.DTN_DISTRICT_AGGREGATED,
             "frequency_id": 1,
@@ -560,6 +651,85 @@ class GroCropClient(object):
             "end_date": end_date
             }
         return self.get_data_points(**params)
+    
+    def get_cash_prices_spot_facility_level(self, crop_id: int, region_id: int, start_date: str = None, end_date: str = None):
+        """Returns facility-level cash prices for a crop and a region."""
+        params = { "source_id": SourceList.DTN,
+                   "frequency_id": 1,
+                   "item_id":  crop_id,
+                   "region_id": region_id,
+                   "metric_id": MetricList.CASH_PRICES_SPOT_DELIVERY_CLOSE,
+                   "start_date": start_date,
+                   "end_date": end_date
+        }
+        return self.get_data_points(**params)
+
+    def get_cash_prices_new_crop_facility_level(self, crop_id: int, region_id: int, start_date: str = None, end_date: str = None):
+        """Returns facility-level cash prices for a crop and a region."""
+        params = { "source_id": SourceList.DTN,
+                   "frequency_id": 1,
+                   "item_id":  crop_id,
+                   "region_id": region_id,
+                   "metric_id": MetricList.CASH_PRICES_NEW_CROP_DELIVERY_CLOSE,
+                   "start_date": start_date,
+                   "end_date": end_date}
+        return self.get_data_points(**params)
+
+
+    def get_basis_prices_spot_facility_level(self, crop_id: int, region_id: int, start_date: str = None, end_date: str = None):
+        """Returns facility-level basis prices for a crop and a region."""
+        params = { "source_id": SourceList.DTN,
+            "frequency_id": 1,
+            "item_id":  crop_id,
+            "region_id": region_id,
+            "metric_id": MetricList.BASIS_PRICES_SPOT_DELIVERY_CLOSE,
+            "start_date": start_date,
+            "end_date": end_date
+           }
+        return self.get_data_points(**params)
+
+    def get_basis_prices_new_crop_facility_level(self, crop_id: int, region_id: int, start_date: str = None, end_date: str = None):
+        """Returns facility-level basis prices for a crop and a region."""
+        params = { "source_id": SourceList.DTN,
+            "frequency_id": 1,
+            "item_id":  crop_id,
+            "region_id": region_id,
+            "metric_id": MetricList.BASIS_PRICES_NEW_CROP_DELIVERY_CLOSE,
+            "start_date": start_date,
+            "end_date": end_date
+            }
+        return self.get_data_points(**params)
+
+    def get_dtn_facilities(self, region_id: int, crop: int, include_lat_longs: bool = False):
+        """Returns DTN facility IDs for a specific region, optionally adding the lat/longs for each facility"""
+        params = {
+            'source_id': SourceList.DTN,
+            'item_id': crop,
+            'region_id': region_id,
+            'metric_id': MetricList.BASIS_PRICES_SPOT_DELIVERY_CLOSE
+        }
+        
+        available_series = self.get_data_series(**params)
+        output_dicts = []
+        region_ids = set()
+        for series in available_series:
+            region_ids.add(series['region_id'])
+        region_ids = sorted(list(region_ids))
+
+        for region_id in region_ids:
+            output_dicts.append({'region_id': region_id})
+
+        if include_lat_longs:
+            region_details = self.lookup('regions', region_ids)
+            for output_dict in output_dicts:
+                region_id_str = str(output_dict['region_id'])
+                if region_id_str in region_details:
+                    output_dict['latitude'] = region_details[region_id_str].get('latitude')
+                    output_dict['longitude'] = region_details[region_id_str].get('longitude')
+                else:
+                    output_dict['latitude'] = output_dict['longitude'] = None
+        
+        return output_dicts
 
     def get_futures_prices_rolling_front_month_settle_national(self, crop_id, start_date = None, end_date = None):
         """Returns future prices for a crop."""
@@ -583,7 +753,8 @@ class GroCropClient(object):
             "start_date": start_date,
             "end_date": end_date
             }
-        return self.get_data_points(**params)      
+        return self.get_data_points(**params)
+      
 
     ##########
     ## Area ##
